SuperStrict

Rem
bbdoc: Asynchronus Tasks
about: This module contains the base types to handle asynchronous tasks. It's inspired by Java's concurrent library.

Create a task by extending TTask and override Execute. Submit the task to an executor. The executor will decide when 
there is a good time to do the task. The returned future can be used to keep track of the task.

Use fscom.AsyncEx to use some example implementations.
End Rem
Module fscom.Async

ModuleInfo "Version: 1.0"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

Rem
bbdoc: This represents a task
about: Tasks can be executed and will return a result or throw an exception.
End Rem
Type TTask
	Rem
	bbdoc: Execute the task
	about: This Method MUST NOT interfere with the TTask-Object.
	So it is safe to call Execute in multiple Threads at once.
	End Rem
	Method Execute:Object() Abstract
End Type


Rem
bbdoc: A Future is an result of an asynchronus task
about: Futures can be obtained upon submitting a task.
They can be used to synchronize threads and keep track of the started tasks.
End Rem
Type TFuture
	Const TIMEOUT_FOREVER:Int = - 1
	
	Rem
	bbdoc: Returns the task this future was created for.
	End Rem
	Method GetTask:TTask() Abstract

	Rem
	bbdoc: Returns true if the task returned a value, threw an exception or was canceled
	End Rem
	Method IsDone:Byte() Abstract
	
	Rem
	bbdoc: Returns true if the task was cancelled and didn't run
	End Rem
	Method IsCanceled:Byte() Abstract
	
	Rem
	bbdoc: Attempt do cancel a task before it was executed
	returns: True if it was successful
	End Rem
	Method Cancel:Byte() Abstract
	
	Rem
	bbdoc: Wait for the task to finish and return the result
	about: This method will block and wait for the task to finish.
	It will throw any exception thrown by the task. 
	It will throw a timeout exception if @timeout was reached before
	the task returned.
	If the task was finished before this method is called, it will 
	imediately return/throw the result
	End Rem
	Method Wait:Object(timeout:Int = TIMEOUT_FOREVER) Abstract
End Type

Rem
bbdoc: An executor that will execute tasks
about: Executors will execute taska that are submitted. The 
tasks may be executed in different threads or in the thread 
Submit or any other method was called.
End Rem
Type TExecutor
	Const TIMEOUT_FOREVER:Int = - 1
	
	Rem
	bbdoc: Submit a task for the executor to execute
	End Rem
	Method Submit:TFuture(task:TTask) Abstract
	
	Rem
	bbdoc: Submit many tasks for the executor to execute
	End Rem	
	Method SubmitAll:TFuture[](tasks:TTask[])
		Local futures:TFuture[tasks.length]
		For Local i:Int = 0 Until tasks.length
			futures[i] = Submit(tasks[i])
		Next
		Return futures
	End Method
	
	Rem
	bbdoc: Executes all Tasks and waits for all of them to be done.
	End Rem
	Method ExecuteAll:TFuture[](tasks:TTask[] , timeout:Int = TIMEOUT_FOREVER) Abstract
	
	Rem
	bbdoc: Executes some Tasks and waits for one to be done and returns its result
	End Rem
	Method ExecuteAny:Object(tasks:TTask[] , timeout:Int = TIMEOUT_FOREVER) Abstract
	
	Rem
	bbdoc: Makes the executor refuse to take any more tasks
	End Rem
	Method Shutdown() Abstract
		
	Rem
	bbdoc: Cancels all submitted tasks and makes the executor refuse to take new tasks
	End Rem
	Method ShutdownNow() Abstract
	
	Rem 
	bbdoc: Waits for shutdown of this executor to be completed
	End Rem
	Method WaitShutdown(timeout:Int = TIMEOUT_FOREVER) Abstract 
		
	Method IsShutdownStarted:Byte() Abstract 
	Method IsShutdownComplete:Byte() Abstract 
End Type