Rem

*** Builproxies ***
Version: 1.2
Author: Francesco Silvani
License: zlib

Usage: buildproxies <sourcefile>

Buildproxy will recrsively run through the files and will 
add a proxy type as sub type for every type that has 
PROXY_METADATA_TAG as metadata.
The proxy type will be stored in a file that will be included
right after the type definition of the parent type.

Do not remove the PROXY_CONST. Buildproxies will recognize it and
will not include files that are already included. So it's save to 
run buldproxies for the same file multiple times.

Run buildproxies after changing any method signature (return type and parameters)
of types you want to have proxy for.

History: 1.1
Added context variable.
Fixed bug with primitives as return values.
History: 1.2
Added constructor call in debug mode to find errors

End Rem

SuperStrict
Framework brl.standardio
Import brl.filesystem
Import "lexer/bmxlexer.bmx"

Const PROXY_CONST$ = "*** PROXY ***"
Const INVOKE_FUNC_NAME$ = "___invoke___"
Const PROXY_FILENAME_SUFFIX$="-proxy.bmx"
Const PROXY_METADATA_TAG$ = "proxy"

?Not debug
Try
?
	TSourceFile.Create(AppArgs[1])
?Not debug
Catch ex:Object
	Print ex.ToString()
	Input ""
End Try 
?

'This type is the referance frame for the Create function.
'Basically Create is a recursive function that will run through every import/include
'and generates the proxy files
Type TSourceFile
	Function Create:TSourceFile(filename$)
		Local this:TSourceFile = New TSourceFile
		this.filename = filename
		
		Local dir:String=CurrentDir()
		ChangeDir ExtractDir(filename)
		
		Try
			this.content = LoadString(StripDir(filename) )
		Catch ex:Object
			Print ex.tostring()
			ChangeDir dir
			Return Null
		End Try
		If this.content = ""
			Return Null
		EndIf
		
		this.newContent=this.content
		
		Print "*** Reading file: "+filename
		this.Run()
		Print "*** Done"
		
		If this.newContent <> this.content
			SaveString this.newContent , filename$
		EndIf
		ChangeDir dir		
		Return this
	End Function 
	
	Field filename$
	Field content$
	Field newContent$
	Field offset%
	Field lexer:TLexer = CreateLexer("bmx")
	Field newFiles:TList=New TList
	
	Method Run()
		lexer.SetSource(content)
		lexer.TidyTokenList()
		
		Local token:TLexerToken = lexer.TokenList.First
		Print token.ToString()
		If token.group = TOKEN_GROUP_COMMENT
			If token.GetValue().StartsWith(PROXY_CONST)
				Print "*** Proxy file. Skip."
				Return
			EndIf
		End If
		
		While token
			If token.group = TOKEN_GROUP_KEYWORD
				Local keyword$ = token.GetValue().ToLower()
				Select keyword
				Case "import","include"
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					
					Create(token.GetValue())
				Case "type"
					Print "*** Type"
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					Local typename:String = token.GetValue()
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					
					If token.group = TOKEN_GROUP_KEYWORD
						If token.GetValue().ToLower() = "abstract"
							token = token.Succ ; Print token.ToString() ; If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						ElseIf token.GetValue().ToLower() = "final"
							Throw "*** Final Types are unable to be proxies"
						EndIf
					EndIf
					
					While token.group=TOKEN_GROUP_IDENT
						If token.GetValue() = PROXY_METADATA_TAG
							Print "*** Generate file"
							GenerateFile(typename,token)

							'Generate Include command
							Local before$ = newContent[..token.lastChar+offset+1]
							Local after$ = newContent[token.lastChar+offset+1..]
							Local between$="~n'"+PROXY_CONST+"~nInclude ~q"+typeName+PROXY_FILENAME_SUFFIX+"~q~n"
							
							token = token.Succ; Print token.ToString(); If token.Succ = Null Then newContent = before + between + after Return
							token = token.Succ; Print token.ToString(); If token.Succ = Null Then newContent = before + between + after Return
							If Not ( token.group=TOKEN_GROUP_COMMENT And token.GetValue().StartsWith(PROXY_CONST)	)
								offset:+between.length
								newContent = before + between + after
							Else
								Print "*** Proxy include found. Skip."
								token = token.Succ; Print token.ToString(); If token.Succ = Null Then Return
								token = token.Succ; Print token.ToString(); If token.Succ = Null Then Return
							EndIf

							Exit 
						EndIf
						token = token.Succ; Print token.ToString(); If token.Succ = Null Then Return
					Wend
				End Select
			EndIf
			
			token = token.Succ; If token Then Print token.ToString()
		Wend
	End Method
	
	Method GenerateFile(typename:String,token:TLexerToken Var)
		Local code$
		code:+ "'"+PROXY_CONST+"~n"
		code:+ "'THIS FILE IS GENERATED. DO NOT EDIT THIS FILE."+"~n"
		code:+ "Type " + typename + "Proxy Extends " + typename+"~n"
		code:+ "	Field "+INVOKE_FUNC_NAME+":Object(this:Object , methodName:String , params:Object[] , metadata:String[], context:Object)"+"~n"
		code:+ "	Field "+INVOKE_FUNC_NAME+"Context:Object"+"~n"
		code:+ "	"+"~n"
		code:+ "	Method Set"+INVOKE_FUNC_NAME+"(pointer:Int)"+"~n"
		code:+ "		"+INVOKE_FUNC_NAME+"=Byte Ptr(pointer)"+"~n"
		code:+ "	End Method"+"~n"
		
		Repeat
			token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
			
			If token.group = TOKEN_GROUP_KEYWORD
				Select token.GetValue().ToLower()
				Case "method"
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					Local methodName$ = token.GetValue()
					Local signature$
					Local metadata$
					Local metadataArray$
					Local paramArray$
					Local hasReturn$
					
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					If token.GetValue() = ":"
						token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						Local returnType$=token.GetValue()
						token = token.Succ ; Print token.ToString() ; If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						If token.GetValue() = "["
							returnType:+"[]"
							token = token.Succ ; Print token.ToString() ; If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
							token = token.Succ ; Print token.ToString() ; If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						EndIf		
						signature:+":"+returnType+"("
						
						If IsPrimitive(returnType)
							hasReturn = "Return " + returnType + "(String("
						Else 
							hasReturn = "Return " + returnType + "("
						EndIf
					ElseIf token.GetValue() = "("
						signature:+ "("
						hasReturn = ""
					Else
						Throw "*** Unexpected Token"
						Return
					EndIf
					
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					
					'If function has parameters
					If Not (token.group = TOKEN_GROUP_SEPARATOR And token.GetValue() = ")")
						BuildParameterList(token,signature,paramArray)
					EndIf
					
					signature:+") "
					
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
										
					'Final methods cannot be overwritten
					If Not (token.group = TOKEN_GROUP_KEYWORD And token.GetValue().ToLower() = "final")
						'Save if the method is abstract to know if there is a method body later
						Local isAbstract:Byte=(token.group = TOKEN_GROUP_KEYWORD And token.GetValue().ToLower()="abstract")
						
						While token.group <> TOKEN_GROUP_EOL
							If token.group = TOKEN_GROUP_IDENT
								metadata:+ token.GetValue()+" "
								metadataArray:+ "~q" + token.GetValue() + "~q, "
							EndIf
							
							token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						Wend
						If metadataArray <> ""
							metadataArray = metadataArray[..metadataArray.length - 2]
							metadata = metadata[..metadata.length-1]
						EndIf 

						If Not IsAbstract
							Repeat
								If token.group = TOKEN_GROUP_KEYWORD
									Select token.GetValue().ToLower()
									Case "end"
										token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
										If token.group = TOKEN_GROUP_KEYWORD And token.GetValue().ToLower() = "method"
											Exit
										EndIf
									Case "end method","endmethod"
										Exit
									End Select
								EndIf
								token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
							Forever
						EndIf
						
						code:+ "~n"
						code:+ "	Method " + methodName + signature
						If metadata
							code:+ " {" + metadata + "}"
						EndIf
						code:+ "~n"
						code:+ "		" + hasReturn + INVOKE_FUNC_NAME + "(Self,~q" + methodName + "~q,"
						If paramArray
							code:+ "[" + paramArray$ + "]"
						Else
							code:+ "Null"
						EndIf
						code:+ ","
						If metadataArray
							code:+ "[" + metadataArray$ + "]"
						Else
							code:+ "Null"
						EndIf
						code:+ ", "+INVOKE_FUNC_NAME+"Context)"
						If hasReturn
							code:+ ")"
							If hasReturn.Find("(String(")>-1
								code:+")"
							EndIf
						EndIf
						code:+ "~n"
						code:+ "	End Method"+"~n"
					EndIf
				Case "end type","endtype"
					Exit
				Case "end"
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					If token.group = TOKEN_GROUP_KEYWORD And token.GetValue().ToLower() = "method"
						Exit
					EndIf
				End Select				
			EndIf 
		Until token.group = TOKEN_GROUP_EOF
		code:+ "End Type"+"~n~n"
		code:+ "'Create dummy instance in debug mode that will yield an error if an abstract method was added"+"~n"
		code:+ "?debug"+"~n"
		code:+ "New "+ typename + "Proxy"+"~n"
		code:+ "?"+"~n"
		
		
		Print code

		SaveString (code , typename + PROXY_FILENAME_SUFFIX)
		Return 
	End Method
	
	Method IsPrimitive:Byte(typeName:String)
		Select typeName
		Case "Byte" , "Short" , "Int" , "Long" , ..
			"Float" , "Double"
			Return True
		Default 
			Return False
		End Select
	End Method
	
	Method BuildParameterList(token:TLexerToken Var , signature:String Var , paramArray:String Var)
		Repeat
			Local paramName$
			Local paramType$
			Local isByReference$=""
			
			If token.group = TOKEN_GROUP_IDENT
				paramName = token.GetValue()
			Else
				Throw "*** Unexpected Token"
				Return
			EndIf
			
			token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
			
			If Not (token.group = TOKEN_GROUP_OPERATOR And token.GetValue() = ":")
				Throw "*** Unexpected Token"
				Return
			EndIf

			token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
			
			If token.group = TOKEN_GROUP_IDENT Or token.group=TOKEN_GROUP_KEYWORD
				paramType = token.GetValue()
			Else
				Throw "*** Unexpected Token"
				Return
			EndIf
			
			token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
			
			'Throw "*** Unexpected Token"
			Repeat
				'Here can be:
				'	,	End of Parameter
				'	[]	Array definition
				'	)	End of Parameterlist
				'	=...	Default value for parameter and constant expression
				'	Var	Destroyed hopes of getting trough a proxy with a call by reference
				If token.group = TOKEN_GROUP_SEPARATOR Or token.group=TOKEN_GROUP_OPERATOR
					Select token.GetValue()
					Case ","
						signature:+ paramName + ":" + paramType + isByReference + ", "					
						If IsPrimitive(paramType)
							paramArray:+ "Object(String("+paramName+")), "
						Else
							paramArray:+ "Object("+paramName+"), "
						EndIf
						Exit 
					Case "["
						paramType:+"[]"
						token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
					Case ")"
						signature:+ paramName + ":" + paramType + isByReference + ", "					
						If IsPrimitive(paramType)
							paramArray:+ "Object(String("+paramName+")), "
						Else
							paramArray:+ "Object("+paramName+"), "
						EndIf
						
						paramArray = paramArray[..paramArray.length - 2]
						signature = signature[..signature.length - 2]
						
						Return 
					Case "="
						Repeat
							token = token.Succ
							Print token.ToString()
							If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
						Until token.group = TOKEN_GROUP_SEPARATOR And (token.GetValue()="," Or token.GetValue()=")")
					Default
						Throw "*** Unexpected Token"
						Return
					EndSelect
				ElseIf token.group= TOKEN_GROUP_KEYWORD And token.GetValue().ToLower()="var"
					isByReference = " Var "
					token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
				Else
					Throw "*** Unexpected Token"
					Return
				EndIf
			Forever
			
			token = token.Succ; Print token.ToString(); If token.Succ = Null Then Throw "*** Unexpected End-Of-File"
		Forever
	End Method
End Type