' bmxlexer.bmx - a lexer for the BlitzMax language

SuperStrict

Import BRL.FileSystem
Import BRL.RamStream ' for Incbin
Import "lexer.bmx"
Incbin "bmxlexer.keywords.ini"


' BMX Lexer Type
Type TBMXLexer Extends TLexer

	' All tokens
	Const TOKEN_IGNORE:Int = - 1 ' ignored token

	' Separators
	Const SEP_COMMA:Int = 1		' ,
	Const SEP_SEMICOLON:Int = 2	' ;
	Const SEP_O_ROUNDBR:Int = 3	' (
	Const SEP_C_ROUNDBR:Int = 4	' )
	Const SEP_O_SQUAREBR:Int = 5	' [
	Const SEP_C_SQUAREBR:Int = 6	' ]

	' Operators
	Const OP_ADD_POS:Int = 10		' +  will be resolved by TidyTokenList()
	Const OP_ADD:Int = 11			' + (binary)
	Const OP_POS:Int = 12			' + (unary)
	Const OP_SUB_NEG:Int = 13		' -  will be resolved by TidyTokenList()
	Const OP_SUB:Int = 14			' - (binary)
	Const OP_NEG:Int = 15			' - (unary)
	Const OP_MUL:Int = 16			' *
	Const OP_DIV:Int = 17			' /
	Const OP_EXP:Int = 18			' ^
	Const OP_BIT_AND:Int = 19		' &
	Const OP_BIT_OR:Int = 20		' |
	Const OP_BIT_XOR_NOT:Int = 21	' ~  will be resolved by TidyTokenList()
	Const OP_BIT_XOR:Int = 22		' ~ (binary)
	Const OP_BIT_NOT:Int = 23		' ~ (unary)
	Const OP_EQUAL:Int = 24			' =
	Const OP_LT:Int = 25			' <
	Const OP_LTEQ:Int = 26			' <=
	Const OP_GT:Int = 27			' >
	Const OP_GTEQ:Int = 28			' >=
	Const OP_NEQ:Int = 29			' <>
	Const OP_MEMBER:Int = 30		' .
	Const OP_SLICE:Int = 31			' ..
	Const OP_COLON:Int = 32			' :
	Const OP_OR:Int = 40			' Or
	Const OP_NEW:Int = 41			' New
	Const OP_MOD:Int = 42			' Mod
	Const OP_SHL:Int = 43			' Shl
	Const OP_SHR:Int = 44			' Shr
	Const OP_SAR:Int = 45			' Sar
	Const OP_NOT:Int = 46			' Not
	Const OP_AND:Int = 47			' And
	Const OP_CHR:Int = 48			' Chr
	Const OP_LEN:Int = 49			' Len
	Const OP_ASC:Int = 50			' Asc
	Const OP_ABS:Int = 51			' Abs
	Const OP_SGN:Int = 52			' Sgn
	Const OP_VARPTR:Int = 53		' VarPtr
	Const OP_SIZEOF:Int = 54		' SizeOf

	' Keywords
	Const KEYWORD_EXTERN:Int = 60
	Const KEYWORD_ENDEXTERN:Int = 61
	Const KEYWORD_FUNCTION:Int = 62
	Const KEYWORD_ENDFUNCTION:Int = 63
	Const KEYWORD_TYPE:Int = 64
	Const KEYWORD_ENDTYPE:Int = 65
	Const KEYWORD_METHOD:Int = 66
	Const KEYWORD_ENDMETHOD:Int = 67
	Const KEYWORD_IF:Int = 68
	Const KEYWORD_THEN:Int = 69
	Const KEYWORD_ELSE:Int = 70
	Const KEYWORD_ELSEIF:Int = 71
	Const KEYWORD_ENDIF:Int = 72
	Const KEYWORD_FOR:Int = 73
	Const KEYWORD_TO:Int = 74
	Const KEYWORD_NEXT:Int = 75
	Const KEYWORD_STEP:Int = 76
	Const KEYWORD_EACHIN:Int = 77
	Const KEYWORD_WHILE:Int = 78
	Const KEYWORD_WEND:Int = 79
	Const KEYWORD_REPEAT:Int = 80
	Const KEYWORD_UNTIL:Int = 81
	Const KEYWORD_FOREVER:Int = 82
	Const KEYWORD_SELECT:Int = 83
	Const KEYWORD_CASE:Int = 84
	Const KEYWORD_DEFAULT:Int = 85
	Const KEYWORD_ENDSELECT:Int = 86
	Const KEYWORD_TRY:Int = 87
	Const KEYWORD_CATCH:Int = 88
	Const KEYWORD_ENDTRY:Int = 89
	Const KEYWORD_END:Int = 90
	Const KEYWORD_RETURN:Int = 91
	Const KEYWORD_CONTINUE:Int = 92
	Const KEYWORD_EXIT:Int = 93
	Const KEYWORD_STRICT:Int = 94
	Const KEYWORD_SUPERSTRICT:Int = 95
	Const KEYWORD_MODULE:Int = 96
	Const KEYWORD_MODULEINFO:Int = 97
	Const KEYWORD_FRAMEWORK:Int = 98
	Const KEYWORD_PUBLIC:Int = 99
	Const KEYWORD_PRIVATE:Int = 100
	Const KEYWORD_LOCAL:Int = 101
	Const KEYWORD_GLOBAL:Int = 102
	Const KEYWORD_CONST:Int = 103
	Const KEYWORD_FIELD:Int = 104
	Const KEYWORD_ABSTRACT:Int = 105
	Const KEYWORD_FINAL:Int = 106
	Const KEYWORD_EXTENDS:Int = 107
	Const KEYWORD_NODEBUG:Int = 108
	Const KEYWORD_BYTE:Int = 109
	Const KEYWORD_SHORT:Int = 110
	Const KEYWORD_INT:Int = 111
	Const KEYWORD_LONG:Int = 112
	Const KEYWORD_FLOAT:Int = 113
	Const KEYWORD_DOUBLE:Int = 114
	Const KEYWORD_VAR:Int = 115
	Const KEYWORD_PTR:Int = 116
	Const KEYWORD_OBJECT:Int = 117
	Const KEYWORD_STRING:Int = 118
	Const KEYWORD_TRUE:Int = 119
	Const KEYWORD_FALSE:Int = 120
	Const KEYWORD_PI:Int = 121
	Const KEYWORD_NULL:Int = 122
	Const KEYWORD_SELF:Int = 123
	Const KEYWORD_SUPER:Int = 124
	Const KEYWORD_RELEASE:Int = 125
	Const KEYWORD_DELETE:Int = 126
	Const KEYWORD_THROW:Int = 127
	Const KEYWORD_ASSERT:Int = 128
	Const KEYWORD_INCBIN:Int = 129
	Const KEYWORD_INCBINPTR:Int = 130
	Const KEYWORD_INCBINLEN:Int = 131
	Const KEYWORD_INCLUDE:Int = 132
	Const KEYWORD_IMPORT:Int = 133
	Const KEYWORD_GOTO:Int = 134
	Const KEYWORD_DEFDATA:Int = 135
	Const KEYWORD_READDATA:Int = 136
	Const KEYWORD_RESTOREDATA:Int = 137
	Const KEYWORD_REM:Int = 138
	Const KEYWORD_ENDREM:Int = 139
	Const KEYWORD_ALIAS:Int = 140

	' Compiler Directives
	Const OTHER_DIRECTIVE:Int = 200 ' unknown compiler directive
	Const OTHER_DIRECTIVE_LINUX:Int = 201			' ?Linux
	Const OTHER_DIRECTIVE_LINUXPPC:Int = 202		' ?LinuxPPC
	Const OTHER_DIRECTIVE_LINUXX86:Int = 203		' ?Linuxx86
	Const OTHER_DIRECTIVE_MACOS:Int = 204			' ?MacOS
	Const OTHER_DIRECTIVE_MACOSPPC:Int = 205		' ?MacOSPPC
	Const OTHER_DIRECTIVE_MACOSX86:Int = 206		' ?MacOSx86
	Const OTHER_DIRECTIVE_WIN32:Int = 207			' ?Win32
	Const OTHER_DIRECTIVE_WIN32X86:Int = 208		' ?Win32x86
	Const OTHER_DIRECTIVE_PPC:Int = 209				' ?PPC
	Const OTHER_DIRECTIVE_X86:Int = 210				' ?x86
	Const OTHER_DIRECTIVE_LITTLEENDIAN:Int = 211	' ?LittleEndian
	Const OTHER_DIRECTIVE_BIGENDIAN:Int = 212		' ?BigEndian
	Const OTHER_DIRECTIVE_DEBUG:Int = 213			' ?Debug
	Const OTHER_DIRECTIVE_END:Int = 214				' ?

	' Other tokens
	Const OTHER_LABEL:Int = 220 ' label

	Method Name:String()
		Return "BlitzMax Lexer"
	EndMethod

	Method FindNextToken:TLexerToken(Pos:Int Var)

		Local State:Int = STATE_START
		Local Token:TLexerToken = New TLexerToken
		Token.LastChar = -1
		Token.Group = TOKEN_GROUP_NONE

		' This function has two key components:
		'   - Finding the characters that make up the next token,
		'   - Storing it as a token (e.g. identifying keywords)
		'
		' FIND THE TOKEN
		' --------------
		' We start in the STATE_START, identifying the first character of the next token
		' Then we proceed to use the relevant STATE_* code to find the end of the token

		While Pos < Source.length - 2

			Pos:+1
			Local Char:Int = Source[Pos]
			Local CharNext:Int = Source[Pos + 1]

			Select State

				' find the next token
				Case STATE_START

					If IsSpace(Char) Then Continue

					' end of line ( CR , LF , CR-LF or LF-CR )
					If Char = CHAR_CR Then
						Token.Group = TOKEN_GROUP_EOL
						Token.FirstChar = Pos
						If CharNext = CHAR_LF Then Pos:+ 1
						Token.LastChar = Pos
						Exit
					EndIf
					If Char = CHAR_LF Then
						Token.Group = TOKEN_GROUP_EOL
						Token.FirstChar = Pos
						If CharNext = CHAR_CR Then Pos:+ 1
						Token.LastChar = Pos
						Exit
					EndIf

					' string literal  "..."
					If Char = CHAR_QUOTE Then
						Token.Group = TOKEN_GROUP_STRING
						Token.FirstChar = Pos
						Token.Prefix = 1
						State = STATE_STRING
					EndIf

					' numeric literal 12 , 1.2 , .12 , $12EF , %101 , 1e2 , 1e-2 , 1.2e3 , 1.2e-3
					If IsNum(Char) Then
						State = STATE_NUM ; Token.FirstChar = Pos
					EndIf
					If Char = CHAR_DOT And IsNum(CharNext) Then ' don't fall for the dot operator
						State = STATE_NUM_FLOAT ; Token.FirstChar = Pos
					EndIf
					If Char = CHAR_DOLLAR Then
						State = STATE_NUM_HEX ; Token.FirstChar = Pos
					EndIf
					If Char = CHAR_PERC Then
						State = STATE_NUM_BIN ; Token.FirstChar = Pos
					EndIf

					' identifier
					If IsAlpha(Char) Then
						State = STATE_IDENT ; Token.FirstChar = Pos
					EndIf

					' label
					If Char = CHAR_HASH Then
						Token.FirstChar = Pos
						State = STATE_LABEL
					EndIf

					' directive
					If Char = CHAR_QUEST Then
						Token.FirstChar = Pos
						State = STATE_DIRECTIVE
					EndIf

					' separator , ; ( ) [ ]
					If IsSepChar(Char) Then
						Token.FirstChar = Pos
						Token.LastChar = Pos
						Token.Group = TOKEN_GROUP_SEPARATOR
						Select Char
							Case CHAR_COMMA ; Token.ID = SEP_COMMA
							Case CHAR_SEMIC ; Token.ID = SEP_SEMICOLON
							Case CHAR_O_ROUNDBR ; Token.ID = SEP_O_ROUNDBR
							Case CHAR_C_ROUNDBR ; Token.ID = SEP_C_ROUNDBR
							Case CHAR_O_SQUAREBR ; Token.ID = SEP_O_SQUAREBR
							Case CHAR_C_SQUAREBR ; Token.ID = SEP_C_SQUAREBR
						EndSelect
						Exit
					EndIf

					' operator
					If IsOpChar(Char) Then
						Token.FirstChar = Pos
						Select Char
							Case CHAR_PLUS Token.ID = OP_ADD
							Case CHAR_MINUS ; Token.ID = OP_SUB_NEG
							Case CHAR_ASTERISK ; Token.ID = OP_MUL
							Case CHAR_SLASH ; Token.ID = OP_DIV
							Case CHAR_CARET ; Token.ID = OP_EXP
							Case CHAR_AMP ; Token.ID = OP_BIT_AND
							Case CHAR_VBAR ; Token.ID = OP_BIT_OR
							Case CHAR_TILDE ; Token.ID = OP_BIT_XOR_NOT
							Case CHAR_EQUAL
								If CharNext = CHAR_LESS Then
									Pos:+ 1 ; Token.ID = OP_LTEQ
								ElseIf CharNext = CHAR_GREATER Then
									Pos:+ 1 ; Token.ID = OP_GTEQ
								Else
									Token.ID = OP_EQUAL
								EndIf
							Case CHAR_LESS
								If CharNext = CHAR_EQUAL Then
									Pos:+ 1 ; Token.ID = OP_LTEQ
								ElseIf CharNext = CHAR_GREATER Then
									Pos:+ 1 ; Token.ID = OP_NEQ
								Else
									Token.ID = OP_LT
								EndIf
							Case CHAR_GREATER
								If CharNext = CHAR_LESS Then
									Pos:+ 1 ; Token.ID = OP_NEQ
								ElseIf CharNext = CHAR_EQUAL Then
									Pos:+ 1 ; Token.ID = OP_GTEQ
								Else
									Token.ID = OP_GT
								EndIf
							Case CHAR_DOT
								If CharNext = CHAR_DOT Then
									Pos:+ 1 ; Token.ID = OP_SLICE
								Else
									Token.ID = OP_MEMBER
								EndIf
							Case CHAR_COLON ; Token.ID = OP_COLON
						EndSelect
						Token.LastChar = Pos
						Token.Group = TOKEN_GROUP_OPERATOR
						Exit
					EndIf


					' line comment
					If Char = CHAR_APOS Then
						State = STATE_COMMENT
						Token.FirstChar = Pos
						Token.Prefix = 1
					EndIf

				Case STATE_STRING

					If (Char = CHAR_CR) Or (Char = CHAR_LF) Then ' EOL in string
						Pos:- 1 ' the EOL-character is not part of the string literal
						Token.LastChar = Pos
						Exit
					EndIf

					If Char = CHAR_QUOTE Then ' string done
						Token.LastChar = Pos
						Token.Postfix = 1
						Exit
					EndIf


				Case STATE_NUM
					If Char = CHAR_DOT Then
						If CharNext <> CHAR_DOT Then ' is float
							State = STATE_NUM_FLOAT
						Else ' found ".."
							Pos:- 1
							Exit
						EndIf
					ElseIf Char = CHAR_LC_E Or Char = CHAR_UC_E Then ' is float in exp notation  1e2
						State = STATE_NUM_EXP
						If CharNext = CHAR_MINUS Then Pos:+1 ' skip "-" of negative exponent 1e-2
					ElseIf Not IsNum(Char) Then ' numeral done
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf

				Case STATE_NUM_FLOAT
					If Char = CHAR_LC_E Or Char = CHAR_UC_E Then ' is float in exp notation  1.2e3
						State = STATE_NUM_EXP
						If CharNext = CHAR_MINUS Then Pos:+1 ' negative exponent  1.2e-3
					ElseIf Not IsNum(Char) Then ' numeral done
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf

				Case STATE_NUM_HEX
					If Not IsNumHex(Char) Then ' numeral done
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf

				Case STATE_NUM_BIN
					If Not IsNumBin(Char) Then ' numeral done
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf

				Case STATE_NUM_EXP
					If Not IsNum(Char) Then ' numeral done
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf


				Case STATE_IDENT
					' check for identifier-prefix: integer%, float#, double!, string$
					If (Char = CHAR_PERC) Or (Char = CHAR_HASH) Or (Char = CHAR_EXCL) Or (Char = CHAR_DOLLAR) Then
						Token.LastChar = Pos
						Token.Postfix = 1
						If (Source[Token.FirstChar] = 82 Or Source[Token.FirstChar] = 114) ..
						And Source[Token.FirstChar..Token.LastChar].ToLower() = "rem" Then ' Rem-Block
							'FIX: Removed check for a later endrem in order to ensure that an unpaired Rem applies to EOF
							'If FindEndRem(Source, Pos) Then
								State = STATE_REMBLOCK
								Token.AddExtData TOKEN_EXT_KEYWORD, "rem", 0
								Token.Prefix = 3
							'Else
							'	Exit
							'EndIf
						Else
							Exit
						EndIf

					ElseIf Not IsAlphaNum(Char) Then ' ident done
						Pos:- 1 ; Token.LastChar = Pos

						If (Source[Token.FirstChar] = 82 Or Source[Token.FirstChar] = 114) ..
						And Source[Token.FirstChar..Token.LastChar + 1].ToLower() = "rem" Then ' Rem-Block
							'FIX: Removed check for a later endrem in order to ensure that an unpaired Rem applies to EOF
							'If FindEndRem(Source, Pos) Then
								State = STATE_REMBLOCK
								Token.AddExtData TOKEN_EXT_KEYWORD, "rem", 0
								Token.Prefix = 3
							'Else
							'	Exit
							'EndIf
						Else
							Exit
						EndIf
					EndIf


				Case STATE_LABEL
					If IsAlpha(Char) Then
						Token.Prefix = Pos - Token.FirstChar
						State = STATE_LABEL_IDENT
					ElseIf Not IsSpace(Char) Then
						Pos:- 1
						Token.LastChar = Token.FirstChar ' only a "#"
						Exit
					EndIf

				Case STATE_LABEL_IDENT
					If Not IsAlphaNum(Char) Then
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf


				Case STATE_DIRECTIVE
					If IsAlpha(Char) Then
						Token.Prefix = Pos - Token.FirstChar
						State = STATE_DIRECTIVE_IDENT
					ElseIf Not IsSpace(Char) Then
						Pos:- 1
						Token.LastChar = Token.FirstChar ' only a "?"
						Exit
					EndIf

				Case STATE_DIRECTIVE_IDENT
					If Not IsAlphaNum(Char) Then
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf


				Case STATE_COMMENT
					Token.LastChar = Pos

					If (Char = CHAR_CR) Or (Char = CHAR_LF) Then
						Pos:- 1 ; Token.LastChar = Pos
						Exit
					EndIf

				Case STATE_REMBLOCK
					Token.LastChar = Pos

					If Char = 77 Or Char = 109 Then ' M   check if last character matches
						If (Source[Pos - 5..Pos + 1].ToLower() = "endrem")
							If IsFirstInLine(Source , Pos - 5) Then
								State = STATE_COMMENT
								Token.AddExtData TOKEN_EXT_KEYWORD, "endrem", Pos - 5 - Token.FirstChar
							EndIf
						ElseIf(Source[Pos - 6..Pos + 1].ToLower() = "end rem") Then
							If IsFirstInLine(Source , Pos - 6) Then
								State = STATE_COMMENT
								Token.AddExtData TOKEN_EXT_KEYWORD, "end", Pos - 6 - Token.FirstChar
								Token.AddExtData TOKEN_EXT_KEYWORD, "rem", Pos - 2 - Token.FirstChar
							EndIf
						EndIf
					EndIf


				Default
					RuntimeError "TBMXLexer.FindNextToken: Abnormal lexer state: " + State

			EndSelect

		Wend

		' TOKENISE
		' --------
		' Now that we have identified the next token, we need to convert it into a 'Proper' token,
		' assigning the correct ID, GROUP, etc.

		If Token.LastChar = - 1 Then Token.LastChar = Min(Pos , Source.length - 1)

		If Token.Group = TOKEN_GROUP_NONE Then
			Select State

				Case STATE_START
					Token.Group = TOKEN_GROUP_EOF
					Token.FirstChar = Source.length
					Token.LastChar = Source.length

				Case STATE_NUM , STATE_NUM_FLOAT , STATE_NUM_HEX , STATE_NUM_BIN , STATE_NUM_EXP
					Token.Group = TOKEN_GROUP_NUMERAL

				Case STATE_IDENT

					Token.ID = 0

					If Keywords Then
						Local en:TStringTableEntry
						en = Keywords.FindEntryByKeyEx(Source[Token.FirstChar + Token.Prefix..Token.LastChar - Token.Postfix + 1])
						If en Then
							If en.Group = 1 Then ' keyword
								Token.ID = en.ID
								Token.Group = TOKEN_GROUP_KEYWORD
							ElseIf en.Group = 2 Then ' operator
								Token.ID = en.ID
								Token.Group = TOKEN_GROUP_OPERATOR
								Token.AddExtData TOKEN_EXT_KEYWORD, en.Key, 0
							EndIf
						EndIf
					EndIf

					If KnownIdents And (Token.ID = 0) Then
						Local en:TStringTableEntry
						en = KnownIdents.FindEntryByKeyEx(Source[Token.FirstChar + Token.Prefix..Token.LastChar - Token.Postfix + 1])
						If en Then
							Token.ID = en.ID
							Token.Group = TOKEN_GROUP_KNOWN_IDENT
						EndIf
					EndIf

					If Token.ID = 0 Then
						Token.Group = TOKEN_GROUP_IDENT
					EndIf


				Case STATE_LABEL , STATE_LABEL_IDENT

					Token.Group = TOKEN_GROUP_OTHER
					Token.ID = OTHER_LABEL

				Case STATE_DIRECTIVE

					Token.Group = TOKEN_GROUP_OTHER
					Token.ID = OTHER_DIRECTIVE_END

				Case STATE_DIRECTIVE_IDENT

					Token.Group = TOKEN_GROUP_OTHER
					Token.ID = OTHER_DIRECTIVE
					If Keywords Then
						Local en:TStringTableEntry
						en = Keywords.FindEntryByKeyEx(Source[Token.FirstChar + Token.Prefix..Token.LastChar - Token.Postfix + 1])
						If en Then
							If en.Group = 3 Then ' directive
								Token.ID = en.ID
							EndIf
						EndIf
					EndIf

				Case STATE_COMMENT , STATE_REMBLOCK

					Token.Group = TOKEN_GROUP_COMMENT

				Default
					RuntimeError "TBMXLexer.FindNextToken: Abnormal tokeniser state: " + State

			EndSelect
		EndIf

		Return Token

	EndMethod

	Method TidyTokenList()
		Local Curr:TLexerToken
		Local Succ:TLexerToken
		Local Temp:TLexerToken

		' Check Rem-Keywords for counterparts that might've just turned up. (UpdateSource)
		Curr = TokenList.First
		While Curr
			If Curr.ID = KEYWORD_REM Then
				If FindEndRem(Source, Curr.LastChar) Then
					UpdateSource Source, Curr.FirstChar, Source.Length - Curr.LastChar
					Return
				EndIf
			EndIf
			Curr = Curr.NextToken()
		Wend

		' First pass:
		' - Merge consecutive comments (EOL's between the comments will be ignored).
		' - Mark line continuations.
		Curr = TokenList.First
		While Curr
			' grab the next token for comparisons
			Succ = Curr.NextToken()

			' If it is a multi-line comment, merge lines together into a single block
			If Curr.Group = TOKEN_GROUP_COMMENT Then
				Temp = Curr
				While Succ
					If Succ.Group = TOKEN_GROUP_COMMENT Then
						Temp = Succ ' store last found comment
					ElseIf Succ.Group = TOKEN_GROUP_EOL Then
						' if it's an EOL, then we will keep going if the next token is a comment
						If succ.nexttoken() And succ.nexttoken().group<>TOKEN_GROUP_COMMENT Then Exit
					Else	
						Exit
					EndIf
					Succ = Succ.NextToken()
				Wend
				If Temp <> Curr Then
					TokenList.MergeTokens Curr, Temp
				EndIf
				Temp = Null
			EndIf

			' If it is a line continuation, merge the line continuation and EOL tokens together as a single
			' EOL token, but mark it to be ignored
			If Curr.ID = OP_SLICE Then
				If Succ.GROUP = TOKEN_GROUP_EOL Then
					Curr.Group = TOKEN_GROUP_EOL
					Curr.ID = TOKEN_IGNORE
					TokenList.MergeTokens Curr, Succ ' extend EOL
				EndIf
			EndIf
			Curr = Curr.NextToken()
		Wend

		' Second pass:
		' - Merge "End ..." tokens. ("End If" into "EndIf", ...)
		' - Merge "Else If" into "ElseIf".
		' - Convert module-indentifiers into strings.
		'    BRL.Basic : <ident><dot><ident> -> <string>
		' - Resolve possibly-unary operators.
		' TODO: Recheck token pairs where the second part has been marked to be ignored.
		GetFirstToken()
		While Current
			Succ = PeekNextToken()
			Select Current.ID
				' Merge "End ..."-Tokens
				Case KEYWORD_END
					Select Succ.ID
						Case KEYWORD_EXTERN ; Current.ID = KEYWORD_ENDEXTERN
						Case KEYWORD_FUNCTION ; Current.ID = KEYWORD_ENDFUNCTION
						Case KEYWORD_TYPE ; Current.ID = KEYWORD_ENDTYPE
						Case KEYWORD_METHOD ; Current.ID = KEYWORD_ENDMETHOD
						Case KEYWORD_IF ; Current.ID = KEYWORD_ENDIF
						Case KEYWORD_SELECT ; Current.ID = KEYWORD_ENDSELECT
						Case KEYWORD_TRY ; Current.ID = KEYWORD_ENDTRY
						Case KEYWORD_REM ; Current.ID = KEYWORD_ENDREM ' For the sake of completeness...
					EndSelect
					If Current.ID <> KEYWORD_END Then
						If Current.NextToken() = Succ Then ' The tokens are immediate neighbours.
							TokenList.MergeTokens Current, Succ
							Succ = Null
						Else
							Succ.ID = TOKEN_IGNORE ' ignore last keyword-part
						EndIf
					EndIf

				' Merge "Else If"
				Case KEYWORD_ELSE
					If Succ.ID = KEYWORD_IF Then
						Current.ID = KEYWORD_ELSEIF
						If Current.NextToken() = PeekNextToken() Then ' The tokens are immediate neighbours.
							TokenList.MergeTokens Current, Succ
							Succ = Null
						Else
							Succ.ID = TOKEN_IGNORE ' ignore last keyword-part
						EndIf
					EndIf

				' Convert module-indentifiers
				Case KEYWORD_FRAMEWORK , KEYWORD_IMPORT , KEYWORD_MODULE
					If (Succ.Group = TOKEN_GROUP_IDENT) ..
					Or (Succ.Group = TOKEN_GROUP_KNOWN_IDENT) ..
					Or (Succ.Group = TOKEN_GROUP_KEYWORD) Then
						Curr = Succ
						Curr.Group = TOKEN_GROUP_STRING
						Succ = Curr.NextToken()

						While Succ And (Succ.ID = OP_MEMBER)
							Succ = Succ.NextToken()
							If Succ And ((Succ.Group = TOKEN_GROUP_IDENT) ..
							Or (Succ.Group = TOKEN_GROUP_KNOWN_IDENT) ..
							Or (Succ.Group = TOKEN_GROUP_KEYWORD) ) Then
								Succ.PrevToken().Remove() ' remove dot
								Curr.LastChar = Succ.LastChar
								Succ = Succ.NextToken()
								Succ.PrevToken().Remove()
							EndIf

						Wend

					EndIf

				' Resolve possibly-unary operators
				Case OP_ADD_POS , OP_SUB_NEG , OP_BIT_XOR_NOT
					Curr = PeekPrevToken()
					If Curr And ((Curr.Group = TOKEN_GROUP_IDENT) ..
					Or (Curr.Group = TOKEN_GROUP_KNOWN_IDENT) ..
					Or (Curr.Group = TOKEN_GROUP_KEYWORD) ..
					Or (Curr.Group = TOKEN_GROUP_NUMERAL) ..
					Or (Curr.Group = TOKEN_GROUP_STRING) ..
					Or (Curr.ID = SEP_C_ROUNDBR) ..
					Or (Curr.ID = SEP_C_SQUAREBR)) Then
						Current.ID:+ 1 ' binary
					Else
						Current.ID:+ 2 ' unary
					EndIf

			EndSelect
			GetNextToken()
		Wend
	EndMethod

	Method GetFirstToken:TLexerToken(Flags:Int = 0)
		Current = TokenList.First
		While Current And ( (Current.ID = TOKEN_IGNORE) Or (Current.Group = TOKEN_GROUP_COMMENT))
			Current = Current.NextToken()
		Wend
		Return Current
	EndMethod

	Method PeekNextToken:TLexerToken(Flags:Int = 0)
		If Not Current Then Return TokenList.First
		Local Temp:TLexerToken = Current
		If Temp Then Temp = Temp.NextToken()
		While Temp And ( (Temp.ID = TOKEN_IGNORE) Or (Temp.Group = TOKEN_GROUP_COMMENT))
			Temp = Temp.NextToken()
		Wend
		Return Temp
	EndMethod

	Method PeekPrevToken:TLexerToken(Flags:Int = 0)
		Local Temp:TLexerToken = Current
		If Temp Then Temp = Temp.PrevToken()
		While Temp And ( (Temp.ID = TOKEN_IGNORE) Or (Temp.Group = TOKEN_GROUP_COMMENT))
			Temp = Temp.PrevToken()
		Wend
		Return Temp
	EndMethod

	Method GetTokenAt:TLexerToken(Pos:Int, Flags:Int = 0)
		Local Temp:TLexerToken = TokenList.First
		While Temp
			If (Temp.FirstChar <= Pos) And (Temp.LastChar >= Pos) Then
				Return Temp
			ElseIf (Temp.LastChar < Pos) Then
				Return Null
			EndIf
			Temp = Temp.NextToken()
		Wend
		If Temp And ( (Temp.ID = TOKEN_IGNORE) Or (Temp.Group = TOKEN_GROUP_COMMENT) ) Then
			Temp = Null
		EndIf
		Return Temp
	EndMethod

	Function Create:TBMXLexer()
		Local lex:TBMXLexer = New TBMXLexer

		' Populate keyword/operator list
		lex.Keywords = TStringTable.CreateFromFile("incbin::bmxlexer.keywords.ini")

		lex.KnownIdents = TStringTable.Create(False , 0) ' ingore case, sort by first char

		Return lex
	EndFunction

EndType

' BMX Lexer Factory
Type TBMXLexerFactory Extends TLexerFactory
	Method CreateLexer:TLexer(ext:String)
		If ext = "bmx" Then
			Return TBMXLexer.Create()
		EndIf
	EndMethod
EndType

New TBMXLexerFactory


' Import commands.txt

Function BMXLexerImportCommands(Lexer:TLexer , url:Object)

	Local s:TStream = ReadStream(url)
	If Not s Then Return

	Lexer.KnownIdents.CreateGroup(url.ToString())

	While Not s.Eof()

		Local command:String = s.ReadLine()

		' remove the help-file url
		If command.Find("|") > -1 Then command = command[..command.Find("|")]

		' remove parameters and returntype
		If command.Find(":") > -1 Then command = command[..command.Find(":")]
		If command.Find("(") > -1 Then command = command[..command.Find("(")]
		If command.Find(" ") > -1 Then command = command[..command.Find(" ")]

		' remove postfix
		If command.Find("$") > -1 Then command = command[..command.Find("$")]
		If command.Find("!") > -1 Then command = command[..command.Find("!")]
		If command.Find("#") > -1 Then command = command[..command.Find("#")]


		If command <> "" Then

			If Lexer.Keywords.FindEntryByKeyEx(command) Then
				'DebugLog " Found Keyword: " + Command
			Else
				Lexer.KnownIdents.AddEntry(command)
			EndIf

		EndIf

	Wend

EndFunction

' Helper Functions
Private

Function IsSpace:Int(c:Int)
	Return (c <= $20) .. ' any control char ...
		And (c <> CHAR_CR) And (c <> CHAR_LF) ' ... except CR and LF
EndFunction

Function IsAlpha:Int(c:Int)
	Return ((c >= $61) And (c <= $7A)) .. ' a - z
		Or ((c >= $41) And (c <= $5A)) .. ' A - Z
		Or (c = $5F) ' _
EndFunction

Function IsNum:Int(c:Int)
	Return (c >= $30) And (c <= $39) ' 0 - 9
EndFunction

Function IsNumHex:Int(c:Int)
	Return IsNum(c) ..
		Or ((c >= $61) And (c <= $66)) .. ' a - f
		Or ((c >= $41) And (c <= $46)) .. ' A - F
EndFunction

Function IsNumBin:Int(c:Int)
	Return (c = $30) Or (c = $31)
EndFunction

Function IsAlphaNum:Int(c:Int)
	Return IsAlpha(c) Or IsNum(c)
EndFunction

Function IsSepChar:Int(c:Int)
	Return (c = CHAR_COMMA) ..
		Or (c = CHAR_SEMIC) ..
		Or (c = CHAR_O_ROUNDBR) ..
		Or (c = CHAR_C_ROUNDBR) ..
		Or (c = CHAR_O_SQUAREBR) ..
		Or (c = CHAR_C_SQUAREBR)
EndFunction

Function IsOpChar:Int(c:Int)
	Return (c = CHAR_PLUS) ..
		Or (c = CHAR_MINUS) ..
		Or (c = CHAR_ASTERISK) ..
		Or (c = CHAR_SLASH) ..
		Or (c = CHAR_CARET) ..
		Or (c = CHAR_AMP) ..
		Or (c = CHAR_VBAR) ..
		Or (c = CHAR_TILDE) ..
		Or (c = CHAR_EQUAL) ..
		Or (c = CHAR_LESS) ..
		Or (c = CHAR_GREATER) ..
		Or (c = CHAR_DOT) ..
		Or (c = CHAR_COLON)
EndFunction

Function IsFirstInLine:Int(Source:String Var, Pos:Int)
	Pos:-1 ' test chars before current
	Repeat
		If (Pos < 0) Or (Source[Pos] = CHAR_CR) Or (Source[Pos] = CHAR_LF) Then
			Return True
		ElseIf Not IsSpace(Source[Pos])
			Return False
		EndIf
		Pos:- 1
	Forever
EndFunction

Function FindEndRem:Int(Source:String Var, Pos:Int)
	Pos:+ 1 ' test chars after current
	Local FirstInLine:Int = False
	Local Temp:String

	While Pos < Source.Length
		Local Char:Int = Source[Pos]

		If (Char = CHAR_CR) Or (Char = CHAR_LF) Then
			FirstInLine = True
		Else
			If FirstInLine Then
				If (Char = $45) Or (Char = $65) Then ' E
					Temp = Source[Pos..Pos+7].ToLower()
					If Temp = "end rem" Then
						Return True
					EndIf
					Temp = Source[Pos..Pos+6].ToLower()
					If Temp = "endrem" Then
						Return True
					EndIf
				EndIf
				If Not IsSpace(Char) Then FirstInLine = False
			EndIf
		EndIf
		Pos:+1
	Wend

	Return False
EndFunction

' for GetNextToken
Const STATE_START:Int = 0

Const STATE_STRING:Int = 1

Const STATE_NUM:Int = 10
Const STATE_NUM_FLOAT:Int = 11
Const STATE_NUM_HEX:Int = 12
Const STATE_NUM_BIN:Int = 13
Const STATE_NUM_EXP:Int = 14

Const STATE_IDENT:Int = 20

Const STATE_LABEL:Int = 30
Const STATE_LABEL_IDENT:Int = 31

Const STATE_DIRECTIVE:Int = 40
Const STATE_DIRECTIVE_IDENT:Int = 41

Const STATE_COMMENT:Int = 50
Const STATE_REMBLOCK:Int = 51


Const CHAR_CR:Int = $0D
Const CHAR_LF:Int = $0A
Const CHAR_EXCL:Int = $21
Const CHAR_QUOTE:Int = $22
Const CHAR_HASH:Int = $23
Const CHAR_DOLLAR:Int = $24
Const CHAR_PERC:Int = $25
Const CHAR_AMP:Int = $26
Const CHAR_APOS:Int = $27
Const CHAR_O_ROUNDBR:Int = $28
Const CHAR_C_ROUNDBR:Int = $29
Const CHAR_ASTERISK:Int = $2A
Const CHAR_PLUS:Int = $2B
Const CHAR_COMMA:Int = $2C
Const CHAR_MINUS:Int = $2D
Const CHAR_DOT:Int = $2E
Const CHAR_SLASH:Int = $2F
Const CHAR_COLON:Int = $3A
Const CHAR_SEMIC:Int = $3B
Const CHAR_LESS:Int = $3C
Const CHAR_EQUAL:Int = $3D
Const CHAR_GREATER:Int = $3E
Const CHAR_QUEST:Int = $3F
Const CHAR_O_SQUAREBR:Int = $5B
Const CHAR_C_SQUAREBR:Int = $5D
Const CHAR_CARET:Int = $5E
Const CHAR_VBAR:Int = $7C
Const CHAR_TILDE:Int = $7E

Const CHAR_UC_E:Int = $45 ' E
Const CHAR_LC_E:Int = $65 ' e



Rem

Local op:String[]
op = ["Or", "New", "Mod", "Shl", "Shr", "Sar", "Not", "And", ..
"Chr", "Len", "Asc", "Abs", "Sgn", "VarPtr", "SizeOf"] ' these are not functions but unary operators

Local kw:String[]
kw = [..
"@Structural", ..
	"Extern", "EndExtern", ..
	"Function", "EndFunction", ..
	"Type", "EndType", ..
	"Method", "EndMethod", ..
	"If", "Then", "Else", "ElseIf", "EndIf", ..
	"For", "To", "Next", "Step", "EachIn", ..
	"While", "Wend", ..
	"Repeat", "Until", "Forever", ..
	"Select", "Case", "Default", "EndSelect", ..
	"Try", "Catch", "EndTry", ..
	"End", "Return", ..
	"Continue", "Exit", ..
"@Informal", ..
	"Strict", "SuperStrict", ..
	"Module", "ModuleInfo", ..
	"Framework", ..
	"Public", "Private", ..
	"Local", "Global", "Const", "Field", ..
	"Abstract", "Final", "Extends", "NoDebug", ..
"@Primitive and Composite Data Types", ..
	"Byte", "Short", "Int", "Long", "Float", "Double", ..
	"Var", "Ptr", ..
"@Internal Data Types", ..
	"Object", "String", ..
"@Internal Identifiers", ..
	"True", "False", "Pi", "Null", "Self", "Super", ..
"@Special Functions", ..
	"Release", "Delete", ..
	"Throw", "Assert", ..
	"Incbin", "IncbinPtr", "IncbinLen", ..
	"Include", "Import", ..
	"Goto", ..
	"DefData", "ReadData", "RestoreData", ..
"@Others", ..
	"Rem", "EndRem", ..
	"Alias"..
]

Local cd:String[]
cd = ["Linux","LinuxPPC","Linuxx86","MacOS","MacOSPPC","MacOSx86","Win32","Win32x86","PPC","x86","LittleEndian","BigEndian","Debug"]


' constants for operator identifiers
n = 40
For s$ = EachIn op

	Print "~tConst OP_" + s.ToUpper() + ":Int = " + n + "~t~t' " + s

	n:+ 1

Next

Print ""


' constants for keyword identifiers
n = 60
For s$ = EachIn kw

	If Not s.Contains("@") Then

		Print "~tConst KEYWORD_" + s.ToUpper() + ":Int = " + n

		n:+ 1

	EndIf

Next

Print ""
Print ""

' constants for compiler directives
n = 201
For s$ = EachIn cd

	Print "~tConst OTHER_DIRECTIVE_" + s.ToUpper() + ":Int = " + n + "~t~t' ?" + s

	n:+ 1

Next

Print ""



Local f:TStream = WriteFile(RequestFile("bmxlexer.keywords.ini", "", True, "bmxlexer.keywords.ini" ))

f.WriteLine "# bmxlexer.keywords.ini - BlitzMax keywords and operators"
f.WriteLine "# Do not edit this file by hand!"
f.WriteLine ""
f.WriteLine "$encoding=latin1"
f.WriteLine "$casesensitive=0"
f.WriteLine "$indexchar=0"
f.WriteLine ""
f.WriteLine "[1,keywords]"


n = 60
For s$ = EachIn kw

	If s.Contains("@") Then
		f.WriteLine "# " + s[1..]
	Else
		f.WriteLine n + "=" + s
		n:+ 1
	EndIf

Next

f.WriteLine ""
f.WriteLine "[2,operators]"

n = 40
For s$ = EachIn op

	f.WriteLine n + "=" + s
	n:+ 1

Next

f.WriteLine ""
f.WriteLine "[3,directives]"

n = 201
For s$ = EachIn cd

	f.WriteLine n + "=" + s
	n:+ 1

Next


CloseFile f

EndRem