SuperStrict

Rem
bbdoc: Proxy Objects
about: Create proxy objects that redirect every method call to one single function. 
Use this to perform security checks on method calls, log method calls, dispatch method calls, etc.

To create a proxy object, you have to add {proxy} to your type. Then have buildproxies run on your source file.
This will generate a proxy type called YourTypeProxy. Use #CreateProxy("YourType",...) to obtain a well prepared object 
of this proxy type.
End Rem
Module fscom.proxy
ModuleInfo "Version: 1.1"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 1.1"
ModuleInfo "History: Added context variable"

Import brl.reflection

Rem
bbdoc: Create an proxy object for a type
End Rem
Function CreateProxy:Object(typeName:String ,targetFunction:Object(this:Object , methodName:String , params:Object[] , metadata:String[], context:Object), context:Object=Null)
	Local typeId:TTypeId = TTypeId.ForName(typeName)
	If Not typeId Throw "Type not found"
	If Not typeId.MetaData(PROXY_METADATA_TAG) Throw "Not a proxy type"
	
	Local proxyTypeId:TTypeId = TTypeId.ForName(typeName + "Proxy")
	If Not proxyTypeId Throw "Proxy type not found"
	Local setInvoke:TMethod=proxyTypeId.FindMethod("Set"+PROXY_INVOKE_FUNC_NAME)
	If Not setInvoke Throw "Proxy type not correctly intialized"
	Local contextField:TField=proxyTypeId.FindField(PROXY_INVOKE_FUNC_NAME+"Context")
	If Not contextField Throw "Proxy type not correctly intialized"
	
	Try
		Local instance:Object = proxyTypeId.NewObject()
		Local pointer:Byte Ptr = targetFunction
		setInvoke.Invoke instance , [Object(String(Int(pointer)))]
		contextField.Set instance , context
		Return instance
	Catch ex:Object
		Throw "Could not create instance"
	End Try  
End Function

'These constants must be synchronized with buildproxies
Const PROXY_CONST$ = "*** PROXY ***"
Const PROXY_INVOKE_FUNC_NAME$ = "___invoke___"
Const PROXY_FILENAME_SUFFIX$="-proxy.bmx"
Const PROXY_METADATA_TAG$="proxy"
