SuperStrict

Rem
bbdoc: RMI / Client
about: This is the client module of RMI. 

In a typical RMI scenario, there will be multiple processes implementing a RMI client
acessing one process that implements a RMI server.

A RMI client is working on a stream and has access to all basic RMI utilities. With RMIPollNetwork you can poll this stream 
so RMI can invoke methods clled by the server on the local objects registered with RMIRegisterObjectServie (see fscom.rmishared).
To invoke remote methods use the proxy objects returned by #RMIGetServiceObject.

If you wish to have a 1 to 1 network relationship you can place one RMI client on each end of a stream and set up the stream by yourself.
Of course you can also use every other kind of stream to invoke methods from or send invocations to 
(but it may be a good idea to use nonblocking methods with streams that can't answer).
End Rem
Module fscom.rmiclient

ModuleInfo "Version: 0.3"
ModuleInfo "Author: Francesco Silvani"
ModuleInfo "License: zlib/libpng"

ModuleInfo "History: 0.3"
ModuleInfo "History: Removed automatic polling since this blocks events."
ModuleInfo "History: 0.2"
ModuleInfo "History: Better exception handling."
ModuleInfo "History: Fixed not returning data in PollNetwork"
ModuleInfo "History: 0.1"
ModuleInfo "History: Removed rmiclientex and inserted here."
ModuleInfo "History: Added local stream polling for receiving invocations from the server"

Import fscom.bufferedstream
Import fscom.rmishared
Import fscom.proxy
Import fscom.asyncex

Rem 
bbdoc: Returns a well prepared object to use as a client.
about: This function will create a proxy object and prepares it to be used as
clientside RMI service.

Every method you call on such an object will call RMIInvokeRemote (see fscom.rmishared) to call a method over RMI.
End Rem
Function RMIGetServiceObject:Object(serviceName:String , localProxyClass:String = "", clientRef:Object=Null)
	If Not localProxyClass
		localProxyClass = serviceName
	EndIf
	Local client:TRMIContext = TRMIContext(clientRef)
	If client = Null client = lastClient
	If client = Null Throw TRMIException.Create("No RMI client started")

	Local ctx:TRMIInvokeRemoteContext = New TRMIInvokeRemoteContext
	ctx.stream = client.stream
	?threaded
		ctx.streamMutex = client.streamMutex
	?
	ctx.serviceName = serviceName
	ctx.executor = client.executor

	Return CreateProxy(localProxyClass , RMIInvokeRemote , ctx)
End Function

Rem
bbdoc: Connect to a host as an RMI client via TCP
returns: An object that identifies this rmi client
End Rem
Function RMIConnect:Object(host:String , port:Short = 55555 , executor:TExecutor=Null)
	Return RMIStartClient(TBufferedStream.CreateTCPClient(host,port), executor)
End Function

Rem
bbdoc: Start the RMI client with a custom stream
returns: An object that identifies this rmi client
End Rem
Function RMIStartClient:Object(stream:TStream , executor:TExecutor=Null)
	If (Not stream) Or (stream.Eof() = True)
		Throw TRMIException.Create("Can't start rmi client.")
	EndIf 
	RMIRegisterObjectService TRMIServiceRegistry.GetRegistry() , "TRMIServiceRegistry"
	
	Local client:TRMIContext = New TRMIContext
	client.stream = stream
	If executor
		client.executor = executor
	Else
		client.executor = New TEagerSingleThreadedExecutor
	EndIf
	?threaded
		client.streamMutex = TMutex.Create()
	?
	
	AddHook RMI_POLL_NETWORK_HOOK , PollNetwork , client
	lastClient=client
	Return client
End Function


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Global lastClient:TRMIContext


Function PollNetwork:Object(hookid:Int,data:Object,context:Object)
	Local client:TRMIContext = TRMIContext(context)
	
	If client
		If client.stream.Eof() = True
			RemoveHook RMI_POLL_NETWORK_HOOK , PollNetwork , client
			client.stream.Close()
			client.stream = Null
			client.executor.Shutdown()
			Throw TRMIException.Create("Connection Lost")
		EndIf
		
		Local i:Int=0
		While (Not client.stream.Eof()) And (RMI_MAX_INVOCATIONS_AT_ONCE<0 Or i<RMI_MAX_INVOCATIONS_AT_ONCE)
			RMIPollStream(client)
			i:+1
		Wend
	EndIf
	Return data
End Function